import logging

logger = logging.getLogger(__name__)

from typing import Callable, Dict, List, Optional

import bdkpython as bdk
from hwilib.common import AddressType as HWIAddressType
from hwilib.descriptor import (
    Descriptor,
    MultisigDescriptor,
    PKHDescriptor,
    PubkeyProvider,
    SHDescriptor,
    TRDescriptor,
    WPKHDescriptor,
    WSHDescriptor,
    parse_descriptor,
)
from hwilib.key import KeyOriginInfo


class ConstDerivationPaths:
    receive = "/0/*"
    change = "/1/*"
    multipath = "/<0;1>/*"


# https://bitcoin.design/guide/glossary/address/
# https://learnmeabitcoin.com/technical/derivation-paths
# https://github.com/bitcoin/bips/blob/master/bip-0380.mediawiki
class AddressType:
    def __init__(
        self,
        name,
        is_multisig,
        key_origin=None,
        bdk_descriptor_secret=None,
        hwi_descriptor_classes=None,
        info_url=None,
        description=None,
        bdk_descriptor=None,
    ) -> None:
        self.name = name
        self.is_multisig = is_multisig
        self.key_origin: Callable = key_origin
        self.bdk_descriptor_secret = bdk_descriptor_secret
        self.info_url = info_url
        self.description = description
        self.bdk_descriptor = bdk_descriptor
        self.hwi_descriptor_classes = hwi_descriptor_classes

    def clone(self):
        return AddressType(
            name=self.name,
            is_multisig=self.is_multisig,
            key_origin=self.key_origin,
            bdk_descriptor_secret=self.bdk_descriptor_secret,
            info_url=self.info_url,
            description=self.description,
            bdk_descriptor=self.bdk_descriptor,
            hwi_descriptor_classes=self.hwi_descriptor_classes,
        )

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return f"AddressType({self.__dict__})"

    def get_bip32_path(self, network: bdk.Network, keychain: bdk.KeychainKind, address_index: int) -> str:
        return f"m/{0 if keychain == bdk.KeychainKind.EXTERNAL else 1}/{address_index}"


class AddressTypes:
    p2pkh = AddressType(
        "Single Sig (Legacy/p2pkh)",
        False,
        key_origin=lambda network: f"m/44h/{0 if network==bdk.Network.BITCOIN else 1}h/0h",
        bdk_descriptor=bdk.Descriptor.new_bip44_public,
        bdk_descriptor_secret=bdk.Descriptor.new_bip44,
        info_url="https://learnmeabitcoin.com/technical/derivation-paths",
        description="Legacy (single sig) addresses that look like 1addresses",
        hwi_descriptor_classes=(PKHDescriptor,),
    )
    p2sh_p2wpkh = AddressType(
        "Single Sig (Nested/p2sh-p2wpkh)",
        False,
        key_origin=lambda network: f"m/49h/{0 if network==bdk.Network.BITCOIN else 1}h/0h",
        bdk_descriptor=bdk.Descriptor.new_bip49_public,
        bdk_descriptor_secret=bdk.Descriptor.new_bip49,
        info_url="https://learnmeabitcoin.com/technical/derivation-paths",
        description="Nested (single sig) addresses that look like 3addresses",
        hwi_descriptor_classes=(SHDescriptor, WPKHDescriptor),
    )
    p2wpkh = AddressType(
        "Single Sig (SegWit/p2wpkh)",
        False,
        key_origin=lambda network: f"m/84h/{0 if network==bdk.Network.BITCOIN else 1}h/0h",
        bdk_descriptor=bdk.Descriptor.new_bip84_public,
        bdk_descriptor_secret=bdk.Descriptor.new_bip84,
        info_url="https://learnmeabitcoin.com/technical/derivation-paths",
        description="SegWit (single sig) addresses that look like bc1addresses",
        hwi_descriptor_classes=(WPKHDescriptor,),
    )
    p2tr = AddressType(
        "Single Sig (Taproot/p2tr)",
        False,
        key_origin=lambda network: f"m/86h/{0 if network==bdk.Network.BITCOIN else 1}h/0h",
        bdk_descriptor=bdk.Descriptor.new_bip86_public,
        bdk_descriptor_secret=bdk.Descriptor.new_bip86,
        info_url="https://github.com/bitcoin/bips/blob/master/bip-0386.mediawiki",
        description="Taproot (single sig) addresses ",
        hwi_descriptor_classes=(TRDescriptor,),
    )
    p2sh_p2wsh = AddressType(
        "Multi Sig (Nested/p2sh-p2wsh)",
        True,
        key_origin=lambda network: f"m/48h/{0 if network==bdk.Network.BITCOIN else 1}h/0h/1h",
        bdk_descriptor_secret=None,
        info_url="https://github.com/bitcoin/bips/blob/master/bip-0048.mediawiki",
        description="Nested (multi sig) addresses that look like 3addresses",
        hwi_descriptor_classes=(SHDescriptor, WSHDescriptor, MultisigDescriptor),
    )
    p2wsh = AddressType(
        "Multi Sig (SegWit/p2wsh)",
        True,
        key_origin=lambda network: f"m/48h/{0 if network==bdk.Network.BITCOIN else 1}h/0h/2h",
        bdk_descriptor_secret=None,
        info_url="https://github.com/bitcoin/bips/blob/master/bip-0048.mediawiki",
        description="SegWit (multi sig) addresses that look like bc1addresses",
        hwi_descriptor_classes=(WSHDescriptor, MultisigDescriptor),
    )


def get_address_type_dicts() -> Dict[str, AddressType]:
    return {k: v for k, v in AddressTypes.__dict__.items() if (not k.startswith("_"))}


def get_all_address_types() -> List[AddressType]:
    return list(get_address_type_dicts().values())


def get_address_types(is_multisig: bool) -> List[AddressType]:
    return [a for a in get_all_address_types() if a.is_multisig == is_multisig]


def get_hwi_address_type(address_type: AddressType) -> HWIAddressType:
    # see https://hwi.readthedocs.io/en/latest/usage/api-usage.html#hwilib.common.AddressType
    if address_type.name in [AddressTypes.p2pkh.name]:
        return HWIAddressType.LEGACY
    if address_type.name in [AddressTypes.p2wpkh.name, AddressTypes.p2wsh.name]:
        return HWIAddressType.WIT
    if address_type.name in [
        AddressTypes.p2sh_p2wpkh.name,
        AddressTypes.p2sh_p2wsh.name,
    ]:
        return HWIAddressType.SH_WIT
    if address_type.name in [AddressTypes.p2tr.name]:
        return HWIAddressType.TAP

    raise ValueError(f"No HWI AddressType could be found for {address_type.name}")


class SimplePubKeyProvider:
    def __init__(
        self,
        xpub: str,
        fingerprint: str,
        key_origin: str,
        derivation_path: str = ConstDerivationPaths.receive,
    ) -> None:
        self.xpub = xpub
        self.fingerprint = self.format_fingerprint(fingerprint)
        # key_origin example: "m/84h/1h/0h"
        self.key_origin = self.format_key_origin(key_origin)
        # derivation_path example "/0/*"
        self.derivation_path = self.format_derivation_path(derivation_path)

    @classmethod
    def format_derivation_path(cls, value):
        assert value.startswith("/")
        return value.replace("'", "h")

    @classmethod
    def format_key_origin(cls, value):
        if value == "m":
            # handle the special case that the key is the highest key without derivation
            return value
        assert value.startswith("m/"), "The value must start with m/"
        return value.replace("'", "h")

    @classmethod
    def is_fingerprint_valid(cls, fingerprint):
        try:
            int(fingerprint, 16)
            return len(fingerprint) == 8
        except ValueError:
            return False

    @classmethod
    def format_fingerprint(cls, value):
        assert cls.is_fingerprint_valid(value)
        return value.upper()

    def clone(self) -> "SimplePubKeyProvider":
        return SimplePubKeyProvider(self.xpub, self.fingerprint, self.key_origin, self.derivation_path)

    def is_testnet(self):
        network_str = self.key_origin.split("/")[2]
        assert network_str.endswith("h")
        network_index = int(network_str.replace("h", ""))
        if network_index == 0:
            return False
        elif network_index == 1:
            return True
        else:
            # https://learnmeabitcoin.com/technical/derivation-paths
            raise ValueError(f"Unknown network/coin type {network_str} in {self.key_origin}")

    @classmethod
    def from_hwi(cls, pubkey_provider: PubkeyProvider) -> "SimplePubKeyProvider":
        return SimplePubKeyProvider(
            xpub=pubkey_provider.pubkey,
            fingerprint=pubkey_provider.origin.fingerprint.hex(),
            key_origin=pubkey_provider.origin.get_derivation_path(),
            derivation_path=pubkey_provider.deriv_path,
        )

    def to_hwi_pubkey_provider(self) -> PubkeyProvider:

        provider = PubkeyProvider(
            origin=KeyOriginInfo.from_string(self.key_origin.replace("m", f"{self.fingerprint}")),
            pubkey=self.xpub,
            deriv_path=self.derivation_path,
        )
        return provider

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"


def _get_descriptor_instances(descriptor: Descriptor) -> List[Descriptor]:
    "Returns the linear chain of chained descriptors . Multiple subdescriptors return an error"
    assert len(descriptor.subdescriptors) <= 1
    if descriptor.subdescriptors:
        result = [descriptor]
        for subdescriptor in descriptor.subdescriptors:
            result += _get_descriptor_instances(subdescriptor)
        return result
    else:
        return [descriptor]


def _find_matching_address_type(
    descriptor_tuple: List[Descriptor], address_types: List[AddressType]
) -> Optional[AddressType]:
    for address_type in address_types:
        if len(descriptor_tuple) == len(address_type.hwi_descriptor_classes) and all(
            isinstance(i, c) for i, c in zip(descriptor_tuple, address_type.hwi_descriptor_classes)
        ):
            return address_type
    return None


class DescriptorInfo:
    def __init__(
        self,
        address_type: AddressType,
        spk_providers: List[SimplePubKeyProvider],
        threshold=1,
    ) -> None:
        self.address_type: AddressType = address_type
        self.spk_providers: List[SimplePubKeyProvider] = spk_providers
        self.threshold: int = threshold

        if not self.address_type.is_multisig:
            assert len(spk_providers) <= 1

    def __repr__(self) -> str:
        return f"{self.__dict__}"

    def get_hwi_descriptor(self, network: bdk.Network):
        # check that the key_origins of the spk_providers are matching the desired output address_type
        common_key_origins = [
            address_type.key_origin(network)
            for address_type in get_all_address_types()
            if address_type.is_multisig
        ]
        for spk_provider in self.spk_providers:
            if spk_provider.key_origin not in common_key_origins:
                logger.warning(f"{spk_provider.key_origin } is not a common multisig key_origin!")

        if self.address_type.is_multisig:
            assert self.address_type.hwi_descriptor_classes[-1] == MultisigDescriptor
            hwi_descriptor = MultisigDescriptor(
                pubkeys=[provider.to_hwi_pubkey_provider() for provider in self.spk_providers],
                thresh=self.threshold,
                is_sorted=True,
            )
        else:
            hwi_descriptor = self.address_type.hwi_descriptor_classes[-1](
                self.spk_providers[0].to_hwi_pubkey_provider()
            )

        for hwi_descriptor_class in reversed(self.address_type.hwi_descriptor_classes[:-1]):
            hwi_descriptor = hwi_descriptor_class(hwi_descriptor)

        return hwi_descriptor

    def get_bdk_descriptor(self, network: bdk.Network):
        return bdk.Descriptor(self.get_hwi_descriptor(network).to_string(), network=network)

    @classmethod
    def from_str(cls, descriptor_str: str) -> "DescriptorInfo":
        hwi_descriptor = parse_descriptor(descriptor_str)

        # first we need to identify the address type
        address_type = _find_matching_address_type(
            _get_descriptor_instances(hwi_descriptor), get_all_address_types()
        )
        if not address_type:
            raise ValueError(f"descriptor {descriptor_str} cannot be matched to known template")

        # get the     pubkey_providers, by "walking to the end of desciptors"
        threshold = 1
        subdescriptor = hwi_descriptor
        for descritptor_class in address_type.hwi_descriptor_classes:
            # just double checking that _find_matching_address_type did its job correctly
            assert isinstance(subdescriptor, descritptor_class)
            subdescriptor = subdescriptor.subdescriptors[0] if subdescriptor.subdescriptors else subdescriptor

        pubkey_providers = subdescriptor.pubkeys
        if isinstance(subdescriptor, MultisigDescriptor):
            # last descriptor is a multisig
            threshold = subdescriptor.thresh

        return DescriptorInfo(
            address_type=address_type,
            spk_providers=[
                SimplePubKeyProvider.from_hwi(pubkey_provider) for pubkey_provider in pubkey_providers
            ],
            threshold=threshold,
        )



import logging
from typing import Dict, List

import bdkpython as bdk

logger = logging.getLogger(__name__)
import base64

from bitcointx import select_chain_params
from bitcointx.core.key import BIP32PathTemplate, KeyStore
from bitcointx.core.psbt import (
    PartiallySignedTransaction as TXPartiallySignedTransaction,
)
from bitcointx.wallet import CCoinExtKey

from .address_types import AddressType, AddressTypes, get_all_address_types
from .device import BaseDevice
from .seed_tools import derive, get_mnemonic_seed


class SoftwareSigner(BaseDevice):
    def __init__(self, mnemonic: str, network: bdk.Network) -> None:
        super().__init__(network=network)
        self.mnemonic = mnemonic

    def get_fingerprint(self) -> str:
        # it doesn't mattrer which AddressTypes i choose, because the fingerprint is identical for all
        address_type = AddressTypes.p2wsh
        xpub, fingerprint = derive(self.mnemonic, address_type.key_origin(self.network), self.network)
        return fingerprint

    def get_xpubs(self) -> Dict[AddressType, str]:
        xpubs = {}
        for address_type in get_all_address_types():
            xpub, fingerprint = derive(self.mnemonic, address_type.key_origin(self.network), self.network)
            xpubs[address_type] = xpub
        return xpubs

    def _extract_derivation_paths(self, input_psbt: bdk.PartiallySignedTransaction) -> List["str"]:
        import json

        psbt_json = json.loads(input_psbt.json_serialize())

        derivation_paths = []

        # Extract input derivation paths from the "bip32_derivation" field in each input
        for input_data in psbt_json["inputs"]:
            bip32_derivation = input_data.get("bip32_derivation")
            if bip32_derivation:
                for derivation_info in bip32_derivation:
                    path = derivation_info[1][-1]  # Get the last element of the path
                    derivation_paths.append(path)

        return derivation_paths

    def sign_psbt(self, input_psbt: bdk.PartiallySignedTransaction) -> bdk.PartiallySignedTransaction:
        # Select network parameters
        network_params = {
            bdk.Network.BITCOIN: "bitcoin",
            bdk.Network.TESTNET: "bitcoin/testnet",
            bdk.Network.REGTEST: "bitcoin/regtest",
            bdk.Network.SIGNET: "bitcoin/signet",
        }
        select_chain_params(network_params.get(self.network, "bitcoin"))

        # Base64-encoded PSBT
        base64_encoded_psbt = input_psbt.serialize()

        # Deserialize the PSBT
        psbt = TXPartiallySignedTransaction.from_base64_or_binary(base64_encoded_psbt)

        ext_key = CCoinExtKey.from_seed(get_mnemonic_seed(self.mnemonic))

        # Create a keystore and add the extended key with the path template
        keystore = KeyStore()
        for path in self._extract_derivation_paths(input_psbt):
            # Define a path template for the key
            keystore.add_key((ext_key, BIP32PathTemplate(path)))

        # Sign the PSBT
        psbt.sign(keystore)

        # Check if the PSBT is finalized (optional, based on your requirements)
        # If not finalized, you might want to handle it differently

        # Serialize the PSBT after signing
        signed_psbt = psbt.serialize()

        # Encode the signed PSBT back to Base64
        signed_base64_psbt = base64.b64encode(signed_psbt).decode()

        return bdk.PartiallySignedTransaction(signed_base64_psbt)

    def sign_message(self, message: str, bip32_path: str) -> str:
        raise NotImplementedError("")

    def display_address(
        self,
        descriptor_str: str,
        keychain: bdk.KeychainKind,
        address_index: int,
        network: bdk.Network,
    ):
        pass



import logging
from typing import Tuple

import bdkpython as bdk
from bitcointx import select_chain_params
from bitcointx.wallet import CCoinExtKey
from mnemonic import Mnemonic

logger = logging.getLogger(__name__)

from .address_types import SimplePubKeyProvider


def key_origin_fits_network(key_origin: str, network: bdk.Network):
    network_str = key_origin.split("/")[2]
    assert network_str.endswith("h")
    network_index = int(network_str.replace("h", ""))

    if network_index == 0:
        return network == bdk.Network.BITCOIN
    elif network_index == 1:
        return network != bdk.Network.BITCOIN
    else:
        raise ValueError(f"Unknown network/coin type {network_str} in {key_origin}")


def get_mnemonic_seed(mnemonic: str):
    mnemo = Mnemonic("english")
    if not mnemo.check(mnemonic):
        raise ValueError("Invalid mnemonic phrase.")
    return mnemo.to_seed(mnemonic)


def derive(mnemonic: str, key_origin: str, network: bdk.Network) -> Tuple[str, str]:
    """returns:
            xpub  (at key_origin)
            fingerprint  (at root)

    Args:
        mnemonic (str): _description_
        key_origin (str): _description_
        network (bdk.Network): _description_

    Raises:
        ValueError: _description_

    Returns:
        Tuple[str, str]: xpub, fingerprint  (where fingerprint is the master fingerprint)
    """
    if not key_origin_fits_network(key_origin, network):
        raise ValueError(f"{key_origin} does not fit the selected network {network}")

    # Select network parameters
    network_params = {
        bdk.Network.BITCOIN: "bitcoin",
        bdk.Network.TESTNET: "bitcoin/testnet",
        bdk.Network.REGTEST: "bitcoin/regtest",
        bdk.Network.SIGNET: "bitcoin/signet",
    }
    select_chain_params(network_params.get(network, "bitcoin"))

    seed_bytes = get_mnemonic_seed(mnemonic)

    # Create a master extended key from the seed
    master_key = CCoinExtKey.from_seed(seed_bytes)

    # Derive the xpub at the specified origin
    derived_key = master_key.derive_path(key_origin)

    # Extract xpub

    xpub = str(derived_key.neuter())

    # Get the fingerprint
    fingerprint = master_key.fingerprint.hex()

    return xpub, fingerprint


def derive_spk_provider(
    mnemonic: str, key_origin: str, network: bdk.Network, derivation_path: str = "/0/*"
) -> SimplePubKeyProvider:
    xpub, fingerprint = derive(mnemonic, key_origin, network)
    return SimplePubKeyProvider(
        xpub=xpub,
        fingerprint=fingerprint,
        key_origin=key_origin,
        derivation_path=derivation_path,
    )



from typing import Dict, Optional, Tuple

import bdkpython as bdk
import hwilib.commands as hwi_commands
from PyQt6.QtWidgets import (
    QComboBox,
    QDialog,
    QLabel,
    QLineEdit,
    QMainWindow,
    QPushButton,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from bitcoin_usb.address_types import AddressType

from .device import USBDevice


class DeviceDialog(QDialog):
    def __init__(self, parent, devices, network):
        super().__init__(parent)
        self.setWindowTitle("Select the detected device")
        self.layout = QVBoxLayout(self)

        # Creating a button for each device
        for device in devices:
            button = QPushButton(f"{device['type']} - {device['model']}", self)
            button.clicked.connect(lambda *args, d=device: self.select_device(d))
            self.layout.addWidget(button)

        self.selected_device = None
        self.network = network

    def select_device(self, device):
        self.selected_device = device
        self.accept()

    def get_selected_device(self):
        return self.selected_device


class InfoDialog(QDialog):
    def __init__(self, message, title="Info"):
        super().__init__()
        self.setWindowTitle(title)

        # Set up the dialog layout
        layout = QVBoxLayout()

        # Add a label with the information message
        message_label = QLabel(message)
        layout.addWidget(message_label)

        # Add an OK button and connect its signal
        ok_button = QPushButton("OK")
        ok_button.clicked.connect(self.accept)
        layout.addWidget(ok_button)

        self.setLayout(layout)


class USBGui:
    def __init__(self, network: bdk.Network, autoselect_if_1_device=False, parent=None) -> None:
        self.autoselect_if_1_device = autoselect_if_1_device
        self.network = network
        self.parent = parent

    def get_device(self) -> Dict:
        devices = hwi_commands.enumerate()
        if not devices:
            InfoDialog("No USB devices found", title="USB Devices").exec()
            return {}
        if len(devices) == 1 and self.autoselect_if_1_device:
            return devices[0]
        else:
            dialog = DeviceDialog(self.parent, devices, self.network)
            if dialog.exec():
                return dialog.get_selected_device()
            else:
                InfoDialog("No device selected", title="USB Devices").exec()
        return {}

    def sign(self, psbt: bdk.PartiallySignedTransaction) -> bdk.PartiallySignedTransaction:
        selected_device = self.get_device()
        if selected_device:
            with USBDevice(selected_device, self.network) as dev:
                return dev.sign_psbt(psbt)
        return None

    def get_fingerprint_and_xpubs(self) -> Optional[Tuple[str, Dict[AddressType, str]]]:
        selected_device = self.get_device()
        if selected_device:
            with USBDevice(selected_device, self.network) as dev:
                return dev.get_fingerprint(), dev.get_xpubs()
        return None

    def get_fingerprint_and_xpub(self, key_origin: str) -> Optional[Tuple[str, str]]:
        selected_device = self.get_device()
        if selected_device:
            with USBDevice(selected_device, self.network) as dev:
                return dev.get_fingerprint(), dev.get_xpub(key_origin)
        return None

    def sign_message(self, message: str, bip32_path: str) -> Optional[str]:
        selected_device = self.get_device()
        if selected_device:
            with USBDevice(selected_device, self.network) as dev:
                return dev.sign_message(message, bip32_path)
        return None

    def display_address(
        self,
        descriptor_str: str,
        keychain: bdk.KeychainKind,
        address_index: int,
    ):
        selected_device = self.get_device()
        if selected_device:
            with USBDevice(selected_device, self.network) as dev:
                dev.display_address(
                    descriptor_str=descriptor_str,
                    keychain=keychain,
                    address_index=address_index,
                    network=self.network,
                )


class MainWindow(QMainWindow):
    def __init__(self, network: bdk.Network):
        super().__init__()
        self.network = network
        self.usb = USBGui(network=network)

        main_widget = QWidget()
        main_widget_layout = QVBoxLayout(main_widget)
        self.setCentralWidget(main_widget)

        self.combo_network = QComboBox(self)
        self.combo_network.addItems([n.name for n in bdk.Network])
        self.combo_network.setCurrentText(self.network.name)
        main_widget_layout.addWidget(self.combo_network)

        # Create a tab widget and set it as the central widget
        tab_widget = QTabWidget(self)
        main_widget_layout.addWidget(tab_widget)

        # Tab 1: XPUBs
        xpubs_tab = QWidget()
        xpubs_layout = QVBoxLayout(xpubs_tab)
        self.button = QPushButton("Get xpubs", xpubs_tab)
        self.button.clicked.connect(self.on_button_clicked)
        xpubs_layout.addWidget(self.button)
        self.xpubs_text_edit = QTextEdit(xpubs_tab)
        self.xpubs_text_edit.setReadOnly(True)
        xpubs_layout.addWidget(self.xpubs_text_edit)
        tab_widget.addTab(xpubs_tab, "XPUBs")

        # Tab 2: PSBT
        psbt_tab = QWidget()
        psbt_layout = QVBoxLayout(psbt_tab)
        self.psbt_text_edit = QTextEdit(psbt_tab)
        self.psbt_text_edit.setPlaceholderText("Paste your PSBT in here")
        psbt_layout.addWidget(self.psbt_text_edit)
        self.psbt_button = QPushButton("Sign PSBT", psbt_tab)
        self.psbt_button.clicked.connect(self.sign)
        psbt_layout.addWidget(self.psbt_button)
        tab_widget.addTab(psbt_tab, "PSBT")

        # Tab 3: Message Signing
        message_tab = QWidget()
        message_layout = QVBoxLayout(message_tab)
        self.message_text_edit = QTextEdit(message_tab)
        self.message_text_edit.setPlaceholderText("Paste your message to be signed")
        message_layout.addWidget(self.message_text_edit)
        self.address_index_line_edit = QLineEdit(message_tab)
        self.address_index_line_edit.setText("m/84h/0h/0h/0/0")
        self.address_index_line_edit.setPlaceholderText("Address index")
        message_layout.addWidget(self.address_index_line_edit)
        self.sign_message_button = QPushButton("Sign Message", message_tab)
        self.sign_message_button.clicked.connect(self.sign_message)
        message_layout.addWidget(self.sign_message_button)
        tab_widget.addTab(message_tab, "Sign Message")

        # Initialize the network selection

        self.combo_network.currentIndexChanged.connect(self.switch_network)
        self.combo_network.setCurrentIndex(0)

    def sign_message(self):
        signed_message = self.usb.sign_message(
            self.message_text_edit.toPlainText(), self.address_index_line_edit.text()
        )
        if signed_message:
            self.message_text_edit.setText(signed_message)

    def sign(self):
        psbt = bdk.PartiallySignedTransaction(self.psbt_text_edit.toPlainText())
        self.psbt_text_edit.setText("")
        signed_psbt = self.usb.sign(psbt)
        if signed_psbt:
            self.psbt_text_edit.setText(signed_psbt.serialize())

    def on_button_clicked(self):
        self.xpubs_text_edit.setText("")
        fingerprint_and_xpus = self.usb.get_fingerprint_and_xpubs()
        if not fingerprint_and_xpus:
            return
        fingerprint, xpubs = fingerprint_and_xpus

        if xpubs:
            txt = "\n".join(
                [
                    f"{str(k)}: [{k.key_origin(self.network).replace('m/',f'{ fingerprint}/')}]  {v}"
                    for k, v in xpubs.items()
                ]
            )

            self.xpubs_text_edit.setText(txt)

    def switch_network(self, idx):
        networks = [n for n in bdk.Network]
        self.network = networks[idx]



import logging
from typing import Optional

import bdkpython as bdk

logger = logging.getLogger(__name__)

from bitcointx.core.psbt import (
    PartiallySignedTransaction as TXPartiallySignedTransaction,
)


class PSBTFinalizer:
    @staticmethod
    def finalize(psbt: bdk.PartiallySignedTransaction) -> Optional[bdk.Transaction]:
        psbt_tx: TXPartiallySignedTransaction = TXPartiallySignedTransaction.from_base64(psbt.serialize())
        # this trys to finalize the tx
        try:
            psbt_tx.extract_transaction()
            if psbt_tx.is_final():
                return bdk.Transaction(psbt_tx.extract_transaction().serialize())
            return None
        except:
            return None



import logging
import threading
from abc import ABC, abstractmethod

import hwilib.commands as hwi_commands
from hwilib.common import Chain
from hwilib.psbt import PSBT

logger = logging.getLogger(__name__)

from typing import Any, Dict

import bdkpython as bdk
from hwilib.descriptor import MultisigDescriptor as HWIMultisigDescriptor

from .address_types import (
    AddressType,
    DescriptorInfo,
    get_all_address_types,
    get_hwi_address_type,
)


class BaseDevice(ABC):
    def __init__(self, network: bdk.Network) -> None:
        self.network = network

    @abstractmethod
    def get_fingerprint(self) -> str:
        pass

    @abstractmethod
    def get_xpubs(self) -> Dict[AddressType, str]:
        pass

    @abstractmethod
    def sign_psbt(self, psbt: bdk.PartiallySignedTransaction) -> bdk.PartiallySignedTransaction:
        pass

    @abstractmethod
    def sign_message(self, message: str, bip32_path: str) -> str:
        pass

    @abstractmethod
    def display_address(
        self,
        descriptor_str: str,
        keychain: bdk.KeychainKind,
        address_index: int,
        network: bdk.Network,
    ) -> str:
        pass


class USBDevice(BaseDevice):
    def __init__(self, selected_device: Dict[str, Any], network: bdk.Network):
        super().__init__(network=network)
        self.selected_device = selected_device
        self.lock = threading.Lock()
        self.client = None

    def bdknetwork_to_chain(self, network: bdk.Network):
        if network.BITCOIN:
            return Chain.MAIN
        if network.REGTEST:
            return Chain.REGTEST
        if network.SIGNET:
            return Chain.SIGNET
        if network.TESTNET:
            return Chain.TEST

    def __enter__(self):
        self.lock.acquire()
        self.client = hwi_commands.get_client(
            device_type=self.selected_device["type"],
            device_path=self.selected_device["path"],
            chain=self.bdknetwork_to_chain(self.network),
        )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.client:
            self.client.close()
        self.lock.release()
        # Handle exceptions if necessary
        if exc_type is not None:
            print(f"An exception occurred: {exc_value}")

    def get_fingerprint(self) -> str:
        assert self.client
        return self.client.get_master_fingerprint().hex()

    def get_xpubs(self) -> Dict[AddressType, str]:
        xpubs = {}
        for address_type in get_all_address_types():
            xpubs[address_type] = self.get_xpub(address_type.key_origin(self.network))
        return xpubs

    def get_xpub(self, key_origin: str) -> str:
        assert self.client
        return self.client.get_pubkey_at_path(key_origin).to_string()

    def sign_psbt(self, psbt: bdk.PartiallySignedTransaction) -> bdk.PartiallySignedTransaction:
        "Returns a signed psbt. However it still needs to be finalized by  a bdk wallet"
        assert self.client
        hwi_psbt = PSBT()
        hwi_psbt.deserialize(psbt.serialize())

        signed_hwi_psbt = self.client.sign_tx(hwi_psbt)

        return bdk.PartiallySignedTransaction(signed_hwi_psbt.serialize())

    def sign_message(self, message: str, bip32_path: str) -> str:
        assert self.client
        return self.client.sign_message(message, bip32_path)

    def display_address(
        self,
        descriptor_str: str,
        keychain: bdk.KeychainKind,
        address_index: int,
        network: bdk.Network,
    ) -> str:
        assert self.client
        desc_infos = DescriptorInfo.from_str(descriptor_str)

        if desc_infos.address_type.is_multisig:
            pubkey_providers = [
                signer_info.to_hwi_pubkey_provider() for signer_info in desc_infos.spk_provider
            ]
            return self.client.display_multisig_address(
                get_hwi_address_type(desc_infos.address_type),
                HWIMultisigDescriptor(
                    pubkeys=pubkey_providers,
                    thresh=desc_infos.threshold,
                    is_sorted=True,
                ),
            )
        else:
            return self.client.display_singlesig_address(
                desc_infos.address_type.get_bip32_path(network, keychain, address_index),
                get_hwi_address_type(desc_infos.address_type),
            )
